// Atomic operations.  These functions are not privileged, but may not be
// available on all architectures.  By default, they operate on 32-bit
// quantities; however, certain functions are also available in a "long"
// version which uses "unsigned long"; these are provided when the
// function is likely to be useful for storing pointers rather than just
// counters.  The long versions are defined arch-x86/atomic.h and
// arch-x64/atomic.h, not here.

#ifndef _LL_ARCH_X86C_ATOMIC_H
#define _LL_ARCH_X86C_ATOMIC_H

#include <lowlevel/types.h>

static inline void ll_atomic_inc(int32_t *val)
{
	if (_LL_SMP)
		asm("lock; incl %0" : "+m" (*val) : : "memory");
	else
		asm("incl %0" : "+m" (*val) : : "memory");
}

static inline void ll_atomic_dec(int32_t *val)
{
	if (_LL_SMP)
		asm("lock; decl %0" : "+m" (*val) : : "memory");
	else
		asm("decl %0" : "+m" (*val) : : "memory");
}

// Increment/Decrement and return non-zero if the new count is zero.

static inline int ll_atomic_inc_and_test(int32_t *val)
{
	uint8_t ret;

	if (_LL_SMP)
		asm("lock; incl %0; setz %b1" : "+m" (*val), "=r" (ret) : : "memory");
	else
		asm("incl %0; setz %b1" : "+m" (*val), "=r" (ret) : : "memory");

	return ret;
}

static inline int ll_atomic_dec_and_test(int32_t *val)
{
	uint8_t ret;

	if (_LL_SMP)
		asm("lock; decl %0; setz %b1" : "+m" (*val), "=r" (ret) : : "memory");
	else
		asm("decl %0; setz %b1" : "+m" (*val), "=r" (ret) : : "memory");

	return ret;
}

static inline int ll_test_and_set(uint32_t *val, int bit)
{
	uint8_t ret;

	if (_LL_SMP)
		asm("lock; btsl %2, %0; setc %b1" : "+m" (*val), "=r" (ret) :
		    "ri" (bit) : "memory");
	else
		asm("btsl %2, %0; setc %b1" : "+m" (*val), "=r" (ret) :
		    "ri" (bit) : "memory");

	return ret;
}

static inline int ll_test_and_clear(uint32_t *val, int bit)
{
	uint8_t ret;

	if (_LL_SMP)
		asm("lock; btrl %2, %0; setc %b1" : "+m" (*val), "=r" (ret) :
		    "ri" (bit) : "memory");
	else
		asm("btrl %2, %0; setc %b1" : "+m" (*val), "=r" (ret) :
		    "ri" (bit) : "memory");

	return ret;
}

static inline int ll_test_and_flip(uint32_t *val, int bit)
{
	uint8_t ret;

	if (_LL_SMP)
		asm("lock; btcl %2, %0; setc %b1" : "+m" (*val), "=r" (ret) :
		    "ri" (bit) : "memory");
	else
		asm("btcl %2, %0; setc %b1" : "+m" (*val), "=r" (ret) :
		    "ri" (bit) : "memory");

	return ret;
}

// Store newval in the pointer if oldval was there before.
// Returns non-zero if the store succeded.

static inline int ll_cmpxchg(uint32_t *val, uint32_t oldval, uint32_t newval)
{
	uint8_t ret;

	if (_LL_SMP)
		asm("lock; cmpxchgl %2, %0; setz %b1" :
		    "+m" (*val), "=r" (ret) : "r" (newval), "a" (oldval) : "memory");
	else
		asm("cmpxchgl %2, %0; setz %b1" :
		    "+m" (*val), "=r" (ret) : "r" (newval), "a" (oldval) : "memory");

	return ret;
}

static inline unsigned long ll_xchg(uint32_t *ptr, uint32_t val)
{
	asm("xchgl %0, %1" : "+r" (val), "+m" (*ptr));
	return val;
}

#endif
