// core/thread.cc -- Scheduler and thread creation/destruction
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#include <util/lock.h>
#include <kern/thread.h>

namespace Lock {
	// OPT: add inline, lockless uncontended case

	// FIXME: Allow a high-priority blocker to steal a lock from
	// a low-priority holder if the holder has not yet run.  This
	// prevents rescheduling every iteration if both threads are
	// repeatedly acquiring and releasing the lock.
	
	void Lock::lock()
	{
		DroppableAutoSpinLockRecIRQ autolock(spinlock);
		
		if (!lockval) {
			lockval = reinterpret_cast<ulong>(curthread);
			return;
		}
		
		assert(lockval != reinterpret_cast<ulong>(curthread));
		
		Threads::ThreadBlocker blocker(curthread);
		waitqueue.block(&blocker);
		
		if (!lockval) {
			lockval = reinterpret_cast<ulong>(curthread);
			waitqueue.unblock(&blocker);
			return;
		}
		
		autolock.unlock();
		curthread->block(&blocker);
		
		// FIXME: interruptible locks
		assert(lockval == reinterpret_cast<ulong>(curthread));
	}
	
	void Lock::unlock()
	{
		AutoSpinLockRecIRQ autolock(spinlock);
		assert(lockval == reinterpret_cast<ulong>(curthread));
		
		if (waitqueue.empty()) {
			lockval = 0;
			return;
		}
		
		Threads::Blocker *b = waitqueue.unblock_one();
		lockval = reinterpret_cast<ulong>(b->thread);
		b->wake();
	}
}
