#ifndef _KERN_TIME_H
#define _KERN_TIME_H

#include <kernel/time.h>
#include <util/spinlock.h>

namespace Time {
	// The clock timer ticks 20 times per second.  This timer is only used
	// for maintaining the wall clock time; user timers, rescheduling, and
	// other purposes are dealt with separately, so this timer only needs
	// to run often enough to avoid overrun problems.
	
	static const int clock_tick_rate = 20;
	
	static const Time clock_tick_increment = {
		0, 1000000000 / clock_tick_rate
	};

	// KTimerEntry overrides the standard TimerEntry in order
	// to bypass the asynchronous Notifier, so that low-level
	// things like rescheduling timers can work.
	
	struct KTimerEntry : public TimerEntry {
	public:
		KTimerEntry()
		{
		}
		
		KTimerEntry(TimerMux *MUX) : TimerEntry(MUX)
		{
		}
	
		typedef void (*functype)(KTimerEntry *kte);
		functype func;
		void *data;
		
		void action()
		{
			func(this);
		}
	};

	// This clock does not change when the wall clock is changed, but
	// rather increases monotonically.  Most timers should use this
	// rather than wall time.

	class MonotonicClock {
		s64 llclocks_per_second;
		s64 llclocks_per_tick;
	
		// The clock's value as of the last tick.
		
		Time last_tick_time;
		s64 last_tick_llclock;
		
		Lock::SpinLock lock;
		
		s32 get_nanos(s32 llclocks)
		{
			return llclocks * 1000000000LL / llclocks_per_second;
		}
		
		KTimerEntry tick_timer;

		static void timer_tick(KTimerEntry *entry);
	
	public:
		#include <servers/core/time/Time/MonotonicClock.h>
		
		MonotonicClock()
		{
			init_iface();
		}
		
		virtual ~MonotonicClock()
		{
		}
		
		void get_time(Time *time);
		void get_resolution(Time *res);
		
		void new_timer(Timer *timer);
		
		void calibrate(u64 llclocks_per_second_init)
		{
			llclocks_per_second = llclocks_per_second_init;
			llclocks_per_tick = llclocks_per_second / clock_tick_rate;
			
			printf("MonotonicClock::calibrate: %lld llclocks per second, %lld per tick\n",
			       llclocks_per_second, llclocks_per_tick);
		}
		
		// This is called every clock tick to update last_tick_time
		// and last_tick_nanos.
		
		void tick();

		void init_tick_timer();
	};
	
	extern MonotonicClock monotonic_clock;
	
	class HWTimer {
	protected:
		Time expiry;
		bool armed;
	
	public:
		#include <servers/core/time/Time/HWTimer.h>
	
		HWTimer()
		{
			init_iface();
			expiry.seconds = 0;
			expiry.nanos = 0;
		}
	
		virtual ~HWTimer()
		{
		}
		
		// If a hardware timer is slow to reprogram, it may ignore
		// arm() if time is later than the currently programmed time.
		// Returns true if the timer action should be run immediately.
		
		virtual void arm(Time new_expiry) = 0;
		virtual void disarm() = 0;
		virtual void calibrate() = 0;
		
		void get_expiry(Time *expiry, uint8_t *armed);

		void set_action(System::Events::Event *event)
		{
			BUG();
		}
	};
	
	extern HWTimer *hw_timer;
	extern TimerMux *monotonic_timers;
	extern KTimerEntry *tick_timer;
	void init();
}

#endif
