// io/irq/i8259.cc -- Intel 8259-compatible IRQ support
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#include <kern/irq.h>
#include <kern/kernel.h>
#include <kern/io.h>
#include <kern/i8259.h>

namespace IRQ {
	using namespace IRQ;

	// This assumes a second i8259 cascaded on IRQ2 of the first i8259,
	// with I/O port bases of 0x20 and 0xa0, as every i8259 I've
	// encountered has been configured that way, even on non-x86
	// platforms.  This can be made more general later if need be.
	//
	// FIXME: For non-x86 platforms, the I/O window base must be added in.
	
	inline void I8259::output_mask(int off)
	{
		ll_out_8(off ? 0xa1 : 0x21, cached_masks[0]);
	}
	
	inline void I8259::mask(u32 irq)
	{
		assert(irq < 16);
	
		int off = irq / 8;
		int bit = irq % 8;
		
		cached_masks[off] |= (1 << bit);
		output_mask(off);
	}
		
	void I8259::unmask(u32 irq)
	{
		assert(irq < 16);

		int off = irq / 8;
		int bit = irq % 8;
		
		cached_masks[off] &= ~(1 << bit);
		output_mask(off);
	}
		
	void I8259::mask_and_ack(u32 irq)
	{
		assert(irq < 16);
		int bit = irq % 8;

		mask(irq);
		
		if (irq < 8) {
			ll_out_8(0x20, 0x60 | bit);
		} else {
			ll_out_8(0xa0, 0x60 | bit);
			ll_out_8(0x20, 0x62);
		}
	}
	
	int I8259::get_pending_irq()
	{
	#if defined(_LL_ARCH_x86) || defined(_LL_ARCH_x64)
		assert(0);
		return -1;
	#else
		ll_out_8(0x20, 0x0b);
		u8 ret = ll_in_8(0x20);
		
		if (!(ret & 0x80))
			return -1;
		
		ret &= 7;
		
		if (ret != 2)
			return ret;
		
		ll_out_8(0xa0, 0x0b);
		ret = ll_in_8(0xa0);
		
		// Shouldn't happen, as even if the device de-asserts the
		// cascaded interrupt, it should wait for EOI.
			
		if (!(ret & 0x80))
			return -1;
		
		return (ret & 7) | 8;
	#endif
	}
		
	void I8259::init()
	{
		// Initialize each 8259 for edge triggered, cascade, 8086 mode. 
		// Interrupt vectors are set to 0x20-0x2f (this only matters on
		// x86/x64).
		
		ll_out_8(0x20, 0x11);
		ll_out_8(0x21, 0x20);
		ll_out_8(0x21, 0x04);
		ll_out_8(0x21, 0x01);
		
		ll_out_8(0xa0, 0x11);
		ll_out_8(0xa1, 0x28);
		ll_out_8(0xa1, 0x02);
		ll_out_8(0xa1, 0x01);
		
		// Mask all IRQs initially.
		
		cached_masks[0] = cached_masks[1] = 0xff;
		ll_out_8(0x21, 0xff);
		ll_out_8(0xa1, 0xff);
		
		bzero(irqslots, sizeof(irqslots));

		for (int i = 0; i < 16; i++)
			irqslots[i].controller = this;
	
		irqs = irqslots;
		num_irqs = 16;
	}

	I8259 i8259;
}
