// mem/addrspace.cc -- System.Mem.AddrSpace
//
// OPT: Special AddrSpaces that only translate/export a linear block of
// another AddrSpace, and don't have individual entries for every page.
//
// OPT: Special VAreas that use their own translation mechanism instead
// of varea->offset, so that filesystem block tables (and similar things)
// don't need to have a VArea per block.
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#include <kern/mem.h>
#include <kern/paging.h>
#include <kern/generic-pagetable.h>
#include <kern/pagetable.h>
#include <kern/pagealloc.h>
#include <kern/generic-pte.h>
#include <kern/compiler.h>
#include <kern/process.h>
#include <kern/thread.h>

extern int roshared_start, roshared_page_end;
extern int rwshared_start, rwshared_page_end;

namespace Mem {
	extern IMappable physmem, anonmem;

	class AddrSpaceFactory {
	public:
		#include <servers/mem/addrspace/Mem/AddrSpaceFactory.h>
		
		AddrSpaceFactory()
		{
			init_iface();
		}
		
		void create(Object *obj)
		{
			*obj = static_cast<IAddrSpace>(*(new AddrSpace));
		}
	};

	class ProcAddrSpaceFactory {
	public:
		#include <servers/mem/addrspace/Mem/ProcAddrSpaceFactory.h>
		
		ProcAddrSpaceFactory()
		{
			init_iface();
		}
		
		void create(Object *obj)
		{
			AddrSpace *as = new ProcAddrSpace;
			Region region;
			MapFlags mf = 0;
			u64 vstart;

			region.start = kvirt_to_phys(&roshared_start);
			region.end = kvirt_to_phys(&roshared_page_end);
			vstart = Arch::roshared_map;
			mf.Fixed = 1;
			mf.access_IDLNS_Read = 1;
			mf.access_IDLNS_Exec = 1;
			
			as->map(physmem, region, &vstart, mf,
			        PTEFlags::protectedmap | PTEFlags::addressonly);
			
			region.start = kvirt_to_phys(&rwshared_start);
			region.end = kvirt_to_phys(&rwshared_page_end);
			vstart = Arch::rwshared_map;
			mf.access_IDLNS_Exec = 0;
			mf.access_IDLNS_Write = 1;
			mf.CopyOnWrite = 1;
			
			as->map(physmem, region, &vstart, mf,
			        PTEFlags::protectedmap | PTEFlags::addressonly);
			
			// Leave the stack no-exec by default.
			region.start = vstart = Arch::stack_bottom;
			region.end = Arch::stack_top;
			mf.CopyOnWrite = 0;
			as->map(anonmem, region, &vstart, mf);
			
			*obj = static_cast<IAddrSpace>(*as);
		}
	};
	
	ProcAddrSpaceFactory real_proc_addrspace_factory;
	Factory proc_addr_space_factory = real_proc_addrspace_factory;

	AddrSpaceFactory real_addrspace_factory;
	Factory addr_space_factory = real_addrspace_factory;

	AddrSpace::AddrSpace(PageTable *ptbl) : mappable(this)
	{
		init_iface();
		is_process = false;
		page_table = ptbl;
		
		if (!ptbl)
			page_table = new PageTableImpl<GenPTE>(false);
		
		cached_free_region = Arch::user_start + Arch::page_size;
	}
	
	ProcAddrSpace::ProcAddrSpace() :
	AddrSpace(new PageTableImpl<Arch::PTE>(true))
	{
		is_process = true;
	}
	
	// This should only be used once during bootup to initialize the
	// kernel's address space with a static initial page table.
	
	ProcAddrSpace::ProcAddrSpace(void *ptbl_toplevel) :
	AddrSpace(new PageTableImpl<Arch::PTE>(ptbl_toplevel))
	{
		// FIXME: set cached_free_region to kernel virtual space
		is_process = true;
	}

	void AddrSpace::get_mappable(IMappable *ma)
	{
		*ma = mappable;
	}

	void AddrSpace::clone(IAddrSpace *addrspace, uint8_t clone_is_real)
	{
		// FIXME: implement
		*addrspace = NULL;
	}
	
	int AddrSpace::handle_fault(ulong vaddr, bool write, bool exec, bool user)
	{
		if (lock.held_by_curthread())
			return false;

		assert(!(write && exec));
		PTEFlags reqflags;

		if (user)
			reqflags.User = 1;

		if (write)
			reqflags.Writeable = 1;
		else if (exec)
			reqflags.Executable = 1;
		else
			reqflags.Readable = 1;
		
		reqflags.Valid = 1;

		try {
			mappable.pagein(page_align(vaddr), reqflags);
		}

		catch (BadPageFault &bpf) {
			return bpf.cause;
		}

		return -1;
	}
	
	bool AddrSpace::check_overlap(Region region, VirtualArea *&va)
	{
		if (region.end < region.start)
			return true;
	
		va = varea_tree.find_nearest(region.start);
		
		if (!va)
			return false;
		
		// If region.start is in an existing region, that region will
		// be returned.

		if (region.end >= va->region().start &&
		    region.start <= va->region().end)
			return true;
		
		// If it returns a region that's greater than region.start, and va
		// itself does not overlap, then prev does not overlap (or else
		// region.start would be in or before prev, and thus prev would
		// have been returned).
		
		// If it returns a region that's less than region.start, we still
		// need to check next, as region.end could be in (or beyond) that
		// region.
		
		if (va->list_node.next != &varea_list) {
			VirtualArea *next =
				va->list_node.next->listentry(VirtualArea, list_node);
		
			if (region.end >= next->region().start &&
		       region.start <= next->region().end)
		   {
		    	va = next;
				return true;
			}
		}

		VirtualArea *prev;
	
		if (va->list_node.prev != &varea_list)
			prev = va->list_node.prev->listentry(VirtualArea, list_node);
		else
			prev = NULL;
	
		if (region.start < va->region().start) {
			assert(!prev || prev->region().end < region.start);
			va = prev;
		}
		
		return false;
	}

	VirtualArea *AddrSpace::split_varea(Region region)
	{
		VirtualArea *start, *mid, *end;

		// check_overlap is used rather than varea_tree.find,
		// so that the first overlapping region can be returned,
		// as most (if not all) callers will need this anyway.
		
		if (!check_overlap(region, start))
			return NULL;
		
		assert(start);
		assert(start->aspace == this);
		assert(start->region().end >= region.start);
		
		if (start->region().start < region.start) {
			// There is a varea that straddles region.start;
			// create a new varea "mid" for the overlapping part.
		
			mid = new VirtualArea;
			
			mid->aspace = this;
			mid->region().start = region.start;
			
			if (region.end > start->region().end)
				mid->region().end = start->region().end;
			else
				mid->region().end = region.end;

			mid->flags = start->flags;
			mid->ma = start->ma;
			mid->offset = start->offset;
			
			if (start->region().end > region.end) {
				// The varea also straddles region.end; create a new
				// varea "end" for the other side of the region.
			
				end = new VirtualArea;

				end->aspace = this;
				end->region().start = region.end + 1;
				end->region().end = start->region().end;

				end->flags = start->flags;
				end->ma = start->ma;
				end->offset = start->offset;
			} else {
				end = NULL;
			}

			start->region().end = region.start - 1;

			varea_tree.add(mid);
			mid->ma->map(mid);
			
			if (end) {
				// Splits have already been done at both ends of the region,
				// so there's no need to look up the ending address.
				
				varea_tree.add(end);
				mid->ma->map(end);
				return mid;
			}
			
			start = mid;
		}
		
		if (start->region().end == region.end)
			return start;
		
		if (start->region().end > region.end)
			end = start;
		else {
			end = varea_tree.find(region.end);
			
			if (!end)
				return start;

			assert(end->aspace == this);
			assert(end->region().start <= region.end);
			assert(end->region().end >= region.end);

			if (end->region().end == region.end)
				return start;
		}
		
		assert(end->region().end > region.end);
		
		// There is a varea that straddles region.start;
		// create a new varea "mid" for the overlapping part.
		
		mid = new VirtualArea;
		
		mid->aspace = this;
		mid->region().start = region.start;
			
		mid->region().start = end->region().start;
		mid->region().end = region.end;

		mid->flags = start->flags;
		mid->ma = start->ma;
		mid->offset = start->offset;
			
		end->region().start = region.end + 1;

		varea_tree.add(mid);
		mid->ma->map(mid);
		
		return start;
	}
	
	bool AddrSpace::get_free_region(ulong len, Region &region,
	                                VirtualArea *&prev)
	{
		assert(page_aligned(len));
		assert(cached_free_region);
	
		region.start = cached_free_region;
		region.end = cached_free_region + len - 1;
		
		if (region.end <= Arch::user_end && !check_overlap(region, prev)) {
			cached_free_region = region.end + 1;
			return true;
		}
		
		for (Util::List *node = &prev->list_node; node != &varea_list;
		     node = node->next)
		{
			VirtualArea *va = node->listentry(VirtualArea, list_node);
			ulong end = Arch::user_end;
			
			if (node->next != &varea_list) {
				VirtualArea *next = node->next->listentry(VirtualArea, list_node);
				end = next->region().start - 1;
			}
			
			assert(end > va->region().end);
			
			if (end - va->region().end >= len) {
				region.start = va->region().end + 1;
				region.end = region.start + len - 1;
				
				assert(page_aligned(region.start));
				cached_free_region = region.end + 1;
				return true;
			}
		}
		
		if (cached_free_region != Arch::user_start + Arch::page_size) {
			cached_free_region = Arch::user_start + Arch::page_size;
			return get_free_region(len, region, prev);
		}
		
		return false;
	}

	// The "mapped" paramater is used to indicate whether the top-level
	// address space has had a mapping established.  If "mapped" is
	// false, but an exception is not thrown, then this method must
	// be called again to propagate the mapping along the aspace chain.
	//
	// FIXME: Between aspace locks, if aspace's mapping is revoked and
	// ma->aspace's mapping changes, a pagein could leak through and cause
	// a page load or a copy-on-write breaking.  This isn't a huge deal
	// (it doesn't affect the correctness of the code or give aspace
	// access to ma->aspace's new mapping), but it's unpleasant, and could
	// have an adverse impact on determinism.  If you have a real-time
	// application that can't tolerate the occasional spurious pagein or
	// copy-on-write breaking, then use an address space that hasn't
	// previously been exposed to recursive mappers.
	
	bool ASpaceMappable::rec_pagein(AddrSpace *aspace, u64 vaddr,
	                                PTEFlags reqflags)
	{
		bool mapped = true;
		
		// aspace->mappable.retain();
		
		while (true) {
			Lock::DroppableAutoLock autolock(aspace->lock);
			VirtualArea *va = aspace->varea_tree.find(vaddr);
			
			if (!va)
				throw BadPageFault(MemoryFault_ns::Cause::Unmapped);

			if ((va->flags & reqflags) != reqflags)
				throw BadPageFault(MemoryFault_ns::Cause::Protected);
			
			if (aspace->map(va, vaddr, reqflags))
				break;

			mapped = false;
			Mappable *ma = va->ma;
			vaddr += va->offset;

			// ma->retain();
			autolock.unlock();
			// aspace->mappable.release();

			if (!ma->is_aspace) {
				ma->pagein(vaddr, reqflags);
				// ma->release();
				break;
			}
			
			aspace = static_cast<ASpaceMappable *>(ma)->aspace;
		}
		
		return mapped;
	}
	
	void ASpaceMappable::pagein(u64 vaddr, PTEFlags reqflags)
	{
		while (!rec_pagein(aspace, vaddr, reqflags));
	}
	
	void AddrSpace::break_copy_on_write(VirtualArea *va, u64 vaddr, u64 phys)
	{
		assert(lock.held_by_curthread());
		assert(rmap_lock.held_by_curthread());

		assert(va->flags.FaultOnWrite);
		assert(va->aspace == this);

		Page *old_page = phys_to_page(phys);

		Region region = { vaddr, vaddr + Arch::page_size - 1 };
	
		// If this is the only reference to the page left, then
		// nothing needs to be copied.  Just clear the COW condition.
		if (is_phys_page(old_page) && old_page->get_refcount() == 1) {
			PTEFlags mask, flags;
			mask.FaultOnWrite = 1;

			page_table->set_flags(region, flags, mask);
			return;
		}
		
		Page *new_page = PageAlloc::alloc(1);

		// FIXME -- highmem
		// OPT: It'd be better to do this without the rmap_lock held,
		// especially if rmap_lock is global rather than per-physpage.
		// I want to keep things simple for now and optimize later,
		// though.

		memcpy(page_to_kvirt(new_page), phys_to_kvirt(phys),
		       Arch::page_size);

		page_table->rmap_table.break_copy_on_write(region.start, new_page);
		new_page->release();
	}

	void ASpaceMappable::get_mapping(u64 vaddr, u64 *phys, PTEFlags *flags)
	{
		aspace->page_table->get_mapping(vaddr, phys, flags);
	}
	
	bool AddrSpace::map(VirtualArea *va, u64 vaddr, PTEFlags reqflags)
	{
		Lock::AutoLock autolock(rmap_lock);
		assert(va->aspace == this);
			
		u64 phys;
		PTEFlags flags;
		va->ma->get_mapping(vaddr + va->offset, &phys, &flags);

		PTEFlags newflags = flags & va->flags;
		newflags.FaultOnWrite = flags.FaultOnWrite | va->flags.FaultOnWrite;
		
		if (!newflags.Valid) {
			assert(va->flags.Valid);
			return false;
		}
		
		if ((newflags & reqflags) != reqflags)
			return false;

		u64 oldphys;
		PTEFlags oldflags;
		page_table->get_mapping(vaddr, &oldphys, &oldflags);
		
		if (oldflags.Valid &&
		    !(reqflags.Writeable && oldflags.FaultOnWrite))
		{
			// If the existing mapping is valid, don't try to map it again.
			// The existing mapping was put there possibly by a race, but
			// more likely because a FaultOnWrite was handled upstream.
			//
			// FaultOnWrite handling is the only type of mapping change that
			// can be done directly; all others must change the varea and do
			// an rmap invalidation instead.  FaultOnWrite is special
			// because we don't want to split vareas for every page that
			// gets its copy-on-write broken.

			assert((oldflags & reqflags) == reqflags);
			assert(!va->flags.FaultOnWrite || oldphys == phys);
			return true;
		}

		if (reqflags.Writeable && oldflags.FaultOnWrite)
		{
			// The FaultOnWrite needs to be handled upstream.
			if (!va->flags.FaultOnWrite)
				return false;
			
			va->aspace->break_copy_on_write(va, vaddr, phys);
		} else {
			assert(!oldflags.Valid);
			PageTable *usptbl = NULL;
			
			if (va->ma->is_aspace) {
				ASpaceMappable *asma = static_cast<ASpaceMappable *>(va->ma);
				usptbl = asma->aspace->page_table;
			}
			
			RMapTable::map(va, usptbl, vaddr, vaddr + va->offset);
			
			RegionWithOffset rwo;
			rwo.start = vaddr;
			rwo.end = vaddr + Arch::page_size - 1;
			rwo.offset = phys;
			
			page_table->map(rwo, newflags);
		}
		
		return true;
	}

	void ASpaceMappable::get_size(u64 *size)
	{
		aspace->get_size(size);
	}
	
	void AddrSpace::map(IMappable ma, Region region, u64 *vstart,
	                    MapFlags mflags, PTEFlags set, PTEFlags clear)
	{
		// FIXME: check alignment for VIPT caches
		// FIXME: Implement the "Replace" map flag
		
		if (mflags.Replace)
			throw_idl(InvalidArgument, 3,
			          countarray("Replace unimplemented"));
		
		Mappable *cma = Mappable::classptr(ma);
		if (!cma) {
			// The given IMappable does not refer to a Mappable
			// of this kernel.

			throw_idl(InvalidArgument, 0, nullarray);
		}
		
		bool fixed = mflags.Fixed;
		
		if (is_process)
			mflags.Fixed = 1;
		
		if (!page_aligned(region.start))
			throw_idl(InvalidArgument, 1, countarray("unaligned start"));
		
		if (!page_aligned(region.end + 1))
			throw_idl(InvalidArgument, 1, countarray("unaligned end"));
		
		Lock::AutoLock autolock(lock);
		Region vregion;
		VirtualArea *prev;
		
		if (*vstart != System::Mem::AddrSpace_ns::unspecified_start) {
			vregion.start = *vstart;
			vregion.end = vregion.start + region.end - region.start;
		
			if (is_process) {
				if (!valid_addr(vregion.start))
					throw_idl(InvalidArgument, 2,
					          countarray("invalid virtual start"));
				
				if (!valid_addr(vregion.end))
					throw_idl(InvalidArgument, 2,
					          countarray("invalid virtual end"));
			}
			
			if (check_overlap(vregion, prev))
				*vstart = System::Mem::AddrSpace_ns::unspecified_start;
		}
		
		if (*vstart == System::Mem::AddrSpace_ns::unspecified_start) {
			if (fixed)
				throw_idl(ResourceBusy, 2, countarray("varea overlap"));
			
			if (!get_free_region(region.end - region.start + 1, vregion, prev))
				throw_idl(OutOfSpace, countarray("out of vspace"));
			
			*vstart = vregion.start;
		}
		
		VirtualArea *newva = new VirtualArea;
		newva->aspace = this;
		newva->region() = vregion;

		newva->flags.Valid = 1;
		newva->flags.User = 1;
		newva->flags.Readable = mflags.access_IDLNS_Read;
		newva->flags.Writeable = mflags.access_IDLNS_Write;
		newva->flags.Executable = mflags.access_IDLNS_Exec;
		newva->flags.FaultOnWrite = mflags.CopyOnWrite;
		newva->flags.raw |= set;
		newva->flags.raw &= ~clear;
		newva->ma = cma;
		newva->offset = region.start - vregion.start;

		varea_tree.add(newva);
		newva->ma->map(newva);
		
		if (prev) {
			prev->list_node.add_front(&newva->list_node);
		} else {
			varea_list.add_front(&newva->list_node);
		}
	}
	
	void AddrSpace::unmap(Region region, bool from_kernel)
	{
		u64 orig_start = region.start;
	
		while (region.start <= region.end) {
			Lock::DroppableAutoLock autolock(lock);
			VirtualArea *va;

			// If check_overlap returns false, then there are no vareas
			// in the specified region, so there's nothing to unmap.
			
			if (!check_overlap(region, va))
				return;
			
			if (va->flags.Protected && !from_kernel) {
				region.start = va->list_node.next->
				               listentry(VirtualArea, list_node)->region().start;

				if (region.start <= orig_start)
					break;
				
				continue;
			}
			
			u64 va_end = va->region().end;
			u64 next_start = 0;
			
			if (va_end > region.end) {
				u64 va_start = va->region().start;
				va->region().start = region.end + 1;
			
				if (va_start < region.start) {
					VirtualArea *newva = new VirtualArea;

					newva->aspace = this;
					newva->region().start = va_start;
					newva->region().end = region.start - 1;
	
					newva->flags = va->flags;
					newva->ma = va->ma;
					newva->offset = va->offset;

					varea_tree.add(newva);
					newva->ma->map(newva);
				}
				
				VirtualArea *nextva =
					va->list_node.next->listentry(VirtualArea, list_node);

				next_start = nextva->region().start;
			} else if (va->region().start < region.start) {
				va->region().end = region.start - 1;
			} else {
				varea_tree.del(va);
				va->ma->unmap(va);
			}

			// This is done after the varea removal, so that new faults
			// don't map things in again.

			// OPT: Skip RMap-based unmapping if nothing maps this aspace.
			// OPT: Push this loop into the RMap code, allowing it to skip
			// empty portions of the tables (as the pagetable code currently
			// does).
			
			while (region.start <= va_end && region.start <= region.end) {
				page_table->rmap_table.unmap(region.start);
				region.start += Arch::page_size;

				if (region.start <= orig_start)
					break;
			}
			
			region.start = next_start;
			
			if (region.start <= orig_start)
				break;
		}
	}
	
	void AddrSpace::set_mapflags(Region region, MapFlags mflags)
	{
		// FIXME: implement
		// Find varea, split if necessary, propagate change to stacked aspaces
	}
	
	void AddrSpace::get_mapflags(Region region, MapFlags *mflags, uint8_t *all_same)
	{
		// FIXME: implement
	}
	
	void AddrSpace::get_mapping(Region region, IMappable *ma, u64 *offset)
	{
		// FIXME: implement
	}
	
	void AddrSpace::get_page_size(u32 *page_size)
	{
		*page_size = Arch::page_size;
	}
	
	void AddrSpace::get_min_align(u32 *min_align)
	{
		*min_align = Arch::page_mapping_min_align;
	}

	void AddrSpace::get_size(u64 *size)
	{
		page_table->get_size(size);
	}
	
	void Mappable::map(VirtualArea *varea)
	{
		mappings_lock.lock_irq();
		mappings.add_back(&varea->mappings_node);
		mappings_lock.unlock_irq();
	}

	void Mappable::unmap(VirtualArea *varea)
	{
		mappings_lock.lock_irq();
		varea->mappings_node.del();
		mappings_lock.unlock_irq();
	}
	
	void PageTable::kill_pte(ulong vaddr, u64 paddr, bool dirty,
	                         bool valid, bool no_release)
	{
		Page *oldpage = phys_to_page(paddr);
		
		if (!is_phys_page(oldpage))
			oldpage = NULL;

		if (is_process && valid) {
			Arch::invalidate_tlb_entry(vaddr);
				
			if (oldpage && dirty &&
			    !ll_test_and_set(&oldpage->flags, PageFlags::bits::Dirty))
			{
				oldpage->retain();
				// Queue page for writeback
			}
		}
		
		if (!no_release && oldpage)
			oldpage->release();
	}
	
	class PhysMem : public Mappable {
	public:
		void get_size(u64 *size)
		{
			if (sizeof(long) == 8)
				*size = 1ULL << (64 - Arch::page_shift);
			else
				*size = 1ULL << (32 - Arch::page_shift);
		}
	
		void pagein(u64 vaddr, PTEFlags reqflags)
		{
		}
		
		void get_mapping(u64 addr, u64 *phys, PTEFlags *flags)
		{
			*phys = addr;
			*flags = 0;
			flags->Valid = 1;
			flags->Readable = 1;
			flags->Writeable = 1;
			flags->Executable = 1;
			flags->User = 1;
			flags->AddressOnly = 1;
		}
	};
	
	PhysMem real_physmem;
	IMappable physmem = real_physmem;

	class AnonMem : public Mappable {
	public:
		void get_size(u64 *size)
		{
			if (sizeof(long) == 8)
				*size = 1ULL << (64 - Arch::page_shift);
			else
				*size = 1ULL << (32 - Arch::page_shift);
		}
	
		void pagein(u64 vaddr, PTEFlags reqflags)
		{
		}
		
		void get_mapping(u64 addr, u64 *phys, PTEFlags *flags)
		{
			Page *page = PageAlloc::alloc(1);
			
			// OPT: Only zero if it was asked for.
			// OPT: Eventually, have separate pagelists for zeroed and
			// unzeroed memory, and a low-priority background thread
			// that zeroes pages and moves them to the zeroed list.
			bzero(page_to_kvirt(page), Arch::page_size);
			
			*phys = page_to_phys(page);
			*flags = 0;
			flags->Valid = 1;
			flags->Readable = 1;
			flags->Writeable = 1;
			flags->Executable = 1;
			flags->User = 1;
		}
	};
	
	AnonMem real_anonmem;
	IMappable anonmem = real_anonmem;
}

#include <servers/mem/addrspace/footer.cc>
